#!/bin/bash

baseline=3997817
enabled=6c1ea7cd
compilecount=10
runcount=10

usage ()
{
    echo "Usage:"
    echo "  $0 [-b buildcount] [-r runcount]"
    exit 0
}

while getopts ":b:r:c:" o; do
    case "${o}" in
        b)
            compilecount=${OPTARG}
            [[ $compilecount =~ ^[0-9]+$ ]] || usage
            ;;
        r)
            runcount=${OPTARG}
            [[ $runcount =~ ^[0-9]+$ ]] || usage
            ;;
        *)
            usage
            ;;
    esac
done
shift $((OPTIND-1))

warn ()
{
    echo "*** $1" >&2
}

checkdir ()
{
    local dir="$1"
    if [ -d "$dir" ]
    then
        if [ -w "$dir" ]
        then
            :
        else
            warn "Can't write into directory $dir"
            exit 1
        fi
    elif [ -f "$dir" ]
    then
        warn "$dir file exists - has to be a directory"
        exit 1
    else
        mkdir "$dir"
    fi
}

checkdir buildoutputs
checkdir runoutputs

dobuilds ()
{
    local file="$1"
    local cdir="$2"
    local i=0
    local istr=""
    while [ $i -lt $compilecount ]
    do
        printf -v istr "%03d" $i
        echo "Build #$istr"
        rm -f "$file.exe"
        make CAKEDIR="../cakeml-$cdir" "$file.exe" 2> buildoutputs/$cdir-$istr-$file
        i=$[ $i + 1 ]
    done
}

doruns ()
{
    local file="$1"
    local cdir="$2"
    local i=0
    local istr=""
    while [ $i -lt $runcount ]
    do
        printf -v istr "%03d" $i
        echo "Run #$istr"
        /usr/bin/time ./"$file.exe" 2> runoutputs/$cdir-$istr-$file
        i=$[ $i + 1 ]
    done
}


dotest ()
{
    local file="$1"
    echo "Testing $file"
    if [ ! -r "bmark_programs/$file.sml" ]
    then
        warn "$file.sml not readable - giving up on it"
        return 1
    fi
    echo "Baseline build"
    dobuilds "$file" "$baseline"
    doruns "$file" "$baseline"
    dobuilds "$file" "$enabled"
    doruns "$file" "$enabled"
}


while read file
do
    dotest "$file"
done < testthese.txt
